#ifndef _PROTEIN_H
#define _PROTEIN_H

//#include "sse.h"
#include <fstream>
#include "myVector.H"
#include <vector>
#include <map>
#include "tree.H"
#include <cstdio>
//#include "pairOfVector.H"
#include <string>
using namespace std;

#define MY_RAND_MAX 2147483647.0

struct neighbourStruct 
{
  int index;
  float dist;
  //    int linkedBit; I won't use this parameter.
};

typedef struct neighbourStruct neighbourType;

struct HHBondStruct
{
  int AA1, AA2;
  int type;
};

typedef struct HHBondStruct HHBondType;

struct HHBondCoordStruct
{
  float NCoord, HCoord;
  int   AAindex1;
  float OCoord, CpCoord;
  int   AAindex2;
};

typedef struct HHBondCoordStruct HHBondCoord;


  



class treeClass;
class nodeClass;
class Sse;

/* 
   sets all distances not within the same SSE to -1. 
*/

void keepDistancesWithinTheSameSSE (vector<vector<float> > &pD);

/*
  set all pairwise distances involving i'th AA to -1.
  So set pD[i][j] = -1.0. for all j, including i. 
  make sure pD matrix is symmetric.
*/
void deleteAllConnections (int i,  vector<vector<float> > & pD);

/* 
   all connections from i..j  will be kept, and others will be set to -1.
   i.e., pD [k][a] = -1.0, where a E [i..j], k E [0..i-1] U [j+1.. numAAs - 1].
   also will maintain the symmetry of the matrix.
*/

void deleteAllConnectionsExceptFrom (int i, int j, vector<vector<float> > & pD);


class Protein 
{


public:
  Protein  (int _index);
  Protein  (float * dofParameters, int numParams, Protein * nativeProtein);
  /* computes the protein out of given parameters */
  
  ~Protein ();
  /*
    similar to readDSSP, but reads the parsed file, which contains
    .) the number of AAs
    .) aa sequence
    .) sse sequence
    .) coord info
  */
  
  void     readParsedDSSPFile  (const char * filename);
    //				vector<vector<float> >  & cACoords);
  void     readDSSPfile        (const char * filename); // does not call 
                                                        //computeEnergy (since
                                                        //loop 
  void     createRandomConfig  (const Protein *);       // does call computeEnergy
  void     createGivenConfig(const vector<float> &dih, 
			     const vector<float> &vec, 
			     const vector<float> &twi, 
			     const vector<float> &len,
			     const Protein * nativeP);

  /*
    sse sequence is updated, to contain only H, E or L, for helix, beta strand 
    and loop.
    The vectors of the protein is computed.
  */
  void     computeSSEVectors (const vector<vector<float> >& cACoords);

  friend   ostream& operator<< (ostream & output, const Protein& p);
  friend   Protein * generateMidConfig(const Protein * p_start, const Protein * p_end);
  friend   bool  proteinConfigParametersFarEnough(const Protein * p_start, const Protein *p_end);
  //  friend   int makeLinkBetweenNodes(/*int *numTrees, treeClass ** globalFirstTree, treeClass ** globalLastTree, int * globalTreeName,  */int numNodes, const int n1, const int n2, const float forwardPathWeight, const float backwardPathWeight);
  friend void makeOneWayLink      (const int n1, const int n2, const float pW);
  friend void printCoordsForPlots (const int index, int notNative);//?
  friend void initNativeStructure (const char * tempStr);
  friend void distanceTest        (Protein ** allNodes, float ** centroCoords);
  friend void displayPath         (Protein *start, Protein *end, 
				   int * pathLength);
  friend void addAnimatedLigand(Protein * protein);
  friend void computeBondDihedralAngles(Protein *) ; // for debugging
  friend void initNativeStructure (int * numberOfNodes, const char * tempStr, 
			  const char * protein_name, 
			  int endNodeIndex) ;
  // above friend added for debugging purposes, can be removed later
  //  void printFirstHelixCentroidCoordinates();
  friend void readNodes       (Protein ** allNodes, int * numNodes,
			       const char filename[] );

  friend void initializeAnn(int numNodes, Protein ** allNodes);
  friend void findNearestNeighbors(int nodeIndex, Protein ** allNodes);
  friend float computeRMSD(const  Protein * p1, const Protein * p2);
  friend void checkPairwiseDistances (Protein * p);
  float    computeEnergy();            // calls _computeEnergy

  /* 
     below is the routine that does the real energy computation.
     computeEnergy calls this.
     segregated like this since we want to be able to compute the energy
     sometimes w/o generating centroid coordinates.
  */
  float    _computeEnergy();

  float    computeHBondEnergyTerm(); 
  /* 
   allocates, and reads c-a coords, centroid coords, C', N and O 
   coordinates.
   set the charge of some amino acids.
   sets loop array, which contain true for every amino acid in a loop.
   sets hydrophobicity array, which contains true for every
        hydrophobic amino acid.
  */
  void     doSetup(const char *path2DataFiles);
  void     setCharges();               /* called by doSetup */
  void     setHydrophobicity();  
  void     setLoops();
  /*
    set up the HH bond list. the file is composed of lines, 
    of the  following format : 
    <type> <AA1> <AA2>.
    type is 1 or 2. 
    type is 1 for the case where the AA with index AAIndex has the 
    N-H and otherAAIndex has the O. 2 for the case where AA with index 
    AAIndex has the and the other AA has the N-H.
  */
  void     setupHHBondsList(string filename); 

  

  //  static map<char,bool> HP_mapping;
  //  static const float minExEnergy;

  void     setHBondDataStructure();
  /*
    generates CA coords, if not generated yet.
    Then prints the pairwise distances, omitting those on the first and last
    loop vectors.
  */
  void     printPairwiseCADistances (const char * filename);
  void     printCACoords            (const char * filename);


  static const float maxExEnergy;
  static  float minEnergy;
  float  getEnergy()   const;                /* returns energy */
  float  getExclusionEnergy()   const;
  void   setEnergy(float iEnergy);
  int    getNumLinks() const;
  bool   checkCollision(); /* 
			      returns true if configuration does not collide
			      with itself, and is low than a 
			      probabilistically generated energy threshold.
			   */
  void createNonRandomConfig(const Protein * original, Protein ** allNodes);
    
  //  const  neighbourType& getNeighbour(int i) const;
  static int    NUMDOF; // num of DOF of the protein, is equal to ... FILL THE BLANK
  //  static double connectionDistanceUpperbound; // the max. distance for trying to connect two protein structures.
  Sse *  getSseElement(int) const;        // access i'th secondary structure element.
  /* i = 0.. numSseElements -1 */
  static int    getNumSseElements();      // returns # sse elements
  static int    getNumAAs();

  /* the below stuff is about getting data for computing  energy. */
      
  void   readRelativecAlphaPositions    (const char filename[]);
  void   readRelativeCentroidPositions  (const char filename[]);
  void   readRelativeCprime_N_OPositions(const char CprimeFilename[], 
					 const char NFilename[],
					 const char OFilename[]); 

  /* calls C-alpha, C', N, O and centroid coordinates calculation routines. */
  void   generateAtomCoords();
  void   generateCAlphaCoords();  
  void   generateCentroidCoords();    /* 
					 for each protein, this should 
					 be run before computing
					 the energy.
				      */
  void   generateCprime_N_OPositions();
  void   generateHCoords();
  
  void   setMyNode (nodeClass * node);     /* sets the myNode variable 
					      to the input parameter */
  
  //  void   setRMSDField(const Protein * nativeP);    
  /*       
	   computes the RMSD from native protein and sets the 
	   RMSDfromNative field accordingly.
  */
 
  void   createObjList();         /* to draw the vectors. */
  void   setPairwiseDistances();  /* 
				     initializes the pD array, which holds
				     the pairwise distances for the AAs 
				     on the same SSE to avoid recomputation.
				  */
  
  static const float 
  minLengthPerAA,maxLengthPerAA; /* 
				    used in generating sse vectors in
				    creating random config.  These are 
				    C-alpha to C-alpha distances. 
				 */
  float distanceSquaredBetweenCAlphas(int i, int j);   // i,j = 0..numAAs-1
  float distanceSquaredBetweenCentroids(int i, int j); // i,j = 0..numAAs-1

  void computeBondDihedralAngles(); /* 
				       computes bond and dihedral angles
				       for the native structure. 
				    */
  void findParams ();
  const vector<float>& getParams() const;

  /* analysis / debug methods */
  void checkBondDihedralAngles();   /* 
				       for debugging, checks whether
				       the non-native config. has the
				       correct bond,dihedral angle values.
				    */
  //  void initRelativeVectorLocations(pairOfVector ** 
  //pairsOfVectors); 

  //void compareRelativeVectorLocations(pairOfVector ** pairsOfVector,
  //int *order,
  //int *numPairsOrdered);
  // in analysis.C.
  /* for each pair of  alpha helices, prints the distance
     as well as the angle.	       */

#if 0
  void   createTotallyElongatedConfig
  (const Protein * myProtein); /* 
				  want to start from totally 
				  elongated config. 
			       */
#endif

  void reportEnergy(); // reports how much of the total energy is due to coulombic, exclusion, 
  //HH interactions

  static void Protein::createObjList(Protein * natP); // for drawing


  
  static bool   isLoop(int AAindex); /*
				       returns true if AAindex'th amino 
				       acid is in loop structure. 
				       AAindex = 0.. numAAs-1
				     */
/* only prints parameters, and to stdout. */
  void          printParameters() const ;
  void          printParameters(const string& nodeFilename) const;
  
  /* fout is already opened, and will not be closed at the end of operation.
     prints parameters of the protein, as well as its energy and rmsd from native. */
  
  void          printParameters(FILE * fout) const ;

  /* prints C-alpha, C' , N O coordinates to file. */
  void          printAtomCoords (char * filename); 
  
  //  float         computeRMSD(const Protein * otherProtein) const;
  
  float         computeDistance(const Protein * otherProtein) const;


  // computes normalized distance in C-space.
  static float  **relcAlphaCoords,  // r, phi and z
      **relCentroidCoords, ** relCprimeCoords, ** relNCoords, **relOCoords;

  //  static float closeEnoughDistance;
  /*
				      distance between two proteins to 
				      consider them as close enough
				      and not to generate middle 
				      configurations anymore.
				    */
  static float MAX_PATH_LENGTH;
  
  //  void       addNeighbour(neighbourType neigh); 
  static map<char,bool>                   hydrophobic;
  static map<int, bool>                   loop;
  static const float minAngleDifference, minLengthDifference;
  
  /* if the parameters between 2 protein config. differ by less than this 
     amount, the two configs are considered as the same. Otherwise floating
     point errors occur.
  */
  //  static const float closeThreshold;
  treeClass *myTree;             /* for tree formation. made public
				    since at least linkNodes want to 
				    access it. */
  nodeClass *myNode;             /* at least initnodecolor wants to
				    access this variable. */
  
  void copyConfig(const Protein * original); 
  /*
    used by genSCData.C and checkpathEnergies,
    the latter is uncalled for the moment
  */

  int getIndex() const;
  //  float getRMSDFromNative() const;



  static const float ENERGY_WEIGHT, LENGTH_WEIGHT;
  static const float angle_epsilon, length_epsilon; 
  // used by createNonRandomConfig. See protein.C for more explanation.
  static const float minSampleClusterSize;
  static vector<char> sseTypeSeq; // 0th element corresponds to the first AA.

  static int modelNumber;
  static float center[3];
  static float objMat[16];
  static int firstAAIndexOfFirstNonLoopVector, // >=0
    lastAAIndexOfLastNonLoopVector_plusOne; // <= numAAs
  static const int uninitializedIndex = -1; // for the index variable.
  static vector<HHBondType> HHBondsList;
  

protected:

  void testInvariant() const throw();

private:

  /* static declarations */
  static const float AApairDistance;
  static int numElements;         /* # secondary structure elements */  
  static vector<char> sequence;
  static vector<bool> charged;    /* true if i'th AA charged. */
  static vector<float> charge;    /* charge of each AA. */
  static vector<int> nonLoopAAIndices; /* indices of AA residues which are 
					  not part of a loop. */
  
  Sse *  SaveStructure(char , int , Point * , char * , char *);

  void     computeLoopVectors();  /*    updates loop vectors so that 
					they point from the end of the 
					previous vector to the beginning 
					of the next vector. If the first 
					vector is a loop, that vector 
					is adjusted so that it points to the
					end of the beginning of the next 
					vector. If the last vector is a loop, 
					it is adjusted so that it starts from
					the end of the previous vector. The 
					vector angles and lengths are also 
					updated accordingly.
				  */

  /* other private variables */
  //  Sse ** elements;
  vector <Sse *> sseVectors; // 0th sse vectors corresponds to the first sse.
  float  energy, coulombicEnergy, exclusionEnergy, HHEnergy, hBondEnergy;
  float  indivChargesCoulombicContribution, alphaHelixCoulombicContribution;
  //  float  rmsdFromNative;
  float **  cAlphaCoords,   /* 
			       allocated by generatecAlphaAtomCoords() 
			       in computeEnergy. deallocated by destructor. 
			       Also to be used to find the distance 
			       between proteins.
			    */
    
    ** centroidCoords,       /* 
				allocated by generateCentroidCoords()
				. deallocated by destructor. 
			     */
    
    ** NCoords, ** OCoords, ** CPrimeCoords;
  
  //  vector<HBondCoord> HBondCoords;
  // allocated with generateHCoords, deallocate 
  // by destructor.
  int numLinks;
  int index;            /* index to reach this protein in allNodes array.*/
  vector<float> params; // normalized parameters. 
  
  /* 
     holds the distance square between the residues which are on the same
     alpha helix (and also beta sheet, maybe, in the future.)
     pD[i][j] = distance; means, the C-alpha distance between 
     ith and jth residues is dist. i and j should be residues on 
     the same alpha helix or beta sheet vector. Otherwise, distance
     is -1.
     
     distance square since the functions I wrote are based on the square 
     of the pairwise centroid distance.
  */
  static vector<vector<float> > pDSqr;//pairwiseDistances sqr;
  

#if 0 
  // below parameters taken directly from receptorclass
  int modelNumber;  
  float center[3];
  float objMat[16];
#endif
  
};





#endif

